/* MEX Version of internal_estimator_family_core                 */
/* use
 *
 *    mex -v -lmwblas -largeArrayDims estimator_family_core.c
 *
 * to compile it */

#include "mex.h"
#include "string.h"
#include "blas.h"
#include <math.h>
#include <omp.h>


void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
    const mxArray *APOSTPDF_SAMPLES = prhs[0];
    const mxArray *LOG_APOSTPDF_VALS    = prhs[1];
    int NUM_OBSERVATIONS = mxGetNumberOfElements(APOSTPDF_SAMPLES);

    double XI1 = mxGetScalar(prhs[2]);
    double XI2 = mxGetScalar(prhs[3]);
    double XI3 = mxGetScalar(prhs[4]);
    double LAMBDA = mxGetScalar(prhs[5]);
    double PHI = mxGetScalar(prhs[6]);
    double PSI = mxGetScalar(prhs[7]);

    int NUM_O,i,j;

    mwSignedIndex M = (mwSignedIndex)mxGetM(mxGetCell(APOSTPDF_SAMPLES,0));
    mwSignedIndex K = (mwSignedIndex)mxGetN(mxGetCell(APOSTPDF_SAMPLES,0));
    mwSignedIndex KM = K*M;

    double *aAPOSTPDF_SAMPLES;
    double *aLOG_APOSTPDF_VALS;
    double *WEIGHTS, *TMP_WEIGHTS;
    double s_WEIGHTS;
    double vmax;
    double *TMP_MATRIX, *TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES;
    double *ARGUMENT_F1;
    double LAMBDA_MINUS_ONE = LAMBDA - 1;

    mxArray *mxRESULT = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
    double *RESULT = mxGetPr(mxRESULT);

    const char *fnames[6] = {"XI1","XI2","XI3","LAMBDA","PHI","PSI"};
    mxArray *mxRESULT_GRAD;
    double *GRAD_XI1, *GRAD_XI2, *GRAD_XI3, *GRAD_LAMBDA, *GRAD_PHI, *GRAD_PSI;
    double *DERIV_F1_DIAG;

    if (nlhs == 2)
    {
        mxRESULT_GRAD = mxCreateStructMatrix(1,1,6,fnames);
        mxArray *mxGRAD_XI1 = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
        mxArray *mxGRAD_XI2 = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
        mxArray *mxGRAD_XI3 = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
        mxArray *mxGRAD_LAMBDA = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
        mxArray *mxGRAD_PHI = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
        mxArray *mxGRAD_PSI = mxCreateDoubleMatrix(M,NUM_OBSERVATIONS,mxREAL);
        GRAD_XI1 = mxGetPr(mxGRAD_XI1);
        GRAD_XI2 = mxGetPr(mxGRAD_XI2);
        GRAD_XI3 = mxGetPr(mxGRAD_XI3);
        GRAD_LAMBDA = mxGetPr(mxGRAD_LAMBDA);
        GRAD_PHI = mxGetPr(mxGRAD_PHI);
        GRAD_PSI = mxGetPr(mxGRAD_PSI);
        mxSetField(mxRESULT_GRAD, 0, "XI1", mxGRAD_XI1);
        mxSetField(mxRESULT_GRAD, 0, "XI2", mxGRAD_XI2);
        mxSetField(mxRESULT_GRAD, 0, "XI3", mxGRAD_XI3);
        mxSetField(mxRESULT_GRAD, 0, "LAMBDA", mxGRAD_LAMBDA);
        mxSetField(mxRESULT_GRAD, 0, "PHI", mxGRAD_PHI);
        mxSetField(mxRESULT_GRAD, 0, "PSI", mxGRAD_PSI);

        /* ARGUMENT_F1 = GRAD_XI1 */
        ARGUMENT_F1 = GRAD_XI1;
    }
    else
    {
        ARGUMENT_F1 = calloc(M*NUM_OBSERVATIONS,sizeof(double));
    }

    char *chn = "N";
    mwSignedIndex iONE = 1;
    double dONE = 1.0;
    double dMINUS_ONE = -1.0;
    double dZERO = 0.0;

    #pragma omp parallel private(aAPOSTPDF_SAMPLES,aLOG_APOSTPDF_VALS,WEIGHTS,TMP_WEIGHTS,vmax,s_WEIGHTS,i,j,TMP_MATRIX,TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES,DERIV_F1_DIAG)
    {

    /* initialisiere Variablen hier, damit sie fuer jeden Kontrollfaden zur Verfuegung stehen */
    if (nlhs == 2)
    {
        DERIV_F1_DIAG = calloc(M,sizeof(double));
    }
    WEIGHTS = calloc(K,sizeof(double));
    TMP_WEIGHTS = calloc(K,sizeof(double));
    TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES = calloc(M*K,sizeof(double));
    TMP_MATRIX = calloc(M*K,sizeof(double));


    /* Es wird versucht, moeglichst viel innerhalb dieser Schleife zu behalten, damit es parallelisiert werden kann */
    #pragma omp for schedule(dynamic) nowait
    for(NUM_O = 0;NUM_O<NUM_OBSERVATIONS;NUM_O++)
    {
        aAPOSTPDF_SAMPLES = mxGetPr(mxGetCell(APOSTPDF_SAMPLES,NUM_O));
        aLOG_APOSTPDF_VALS = mxGetPr(mxGetCell(LOG_APOSTPDF_VALS,NUM_O));

        /*****************************************************************/
        /** 1. CALC ESTIMATE                                            **/
        /*****************************************************************/

        /* WEIGHTS = LOG_APOSTPDF_VALS{NUM_O}*(FPARAMS.LAMBDA - 1); */
        /* WEIGHTS = WEIGHTS - max(WEIGHTS); */
        /* WEIGHTS = exp(WEIGHTS); */
        
        memcpy(WEIGHTS, aLOG_APOSTPDF_VALS, K*sizeof(double));
        dscal(&K, &LAMBDA_MINUS_ONE, WEIGHTS, &iONE);

        s_WEIGHTS = 0;
        for(i = 0;i < K;i++)
        {
            WEIGHTS[i] = exp(WEIGHTS[i]);
            s_WEIGHTS += WEIGHTS[i];
        }
        s_WEIGHTS = 1/s_WEIGHTS;
        dscal(&K, &s_WEIGHTS, WEIGHTS, &iONE);
        

        /* FVAL_F2 = f2(APOSTPDF_SAMPLES{NUM_O}, FPARAMS.XI2, FPARAMS.PSI) */
        for(i = 0;i < K*M; i++)
        {
            TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES[i] = exp(PSI*aAPOSTPDF_SAMPLES[i]);
        }
        memcpy(TMP_MATRIX, TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES, K*M*sizeof(double));
        dscal(&KM, &XI3, TMP_MATRIX, &iONE);
        daxpy(&KM, &XI2, aAPOSTPDF_SAMPLES, &iONE, TMP_MATRIX, &iONE);

        /* ARGUMENT_F1 = FVAL_F2 * WEIGHTS' / sum(WEIGHTS); */
        dgemv(chn, &M, &K, &dONE, TMP_MATRIX, &M, WEIGHTS, &iONE, &dZERO, ARGUMENT_F1+NUM_O*M, &iONE );

        /* FVAL_F1 = f1( ARGUMENT_F1, FPARAMS.XI1, FPARAMS.PHI); */
        for(i = M*NUM_O;i < M*(NUM_O+1); i++)
        {
            if (PHI != 0.0)
            {
                RESULT[i] = XI1*ARGUMENT_F1[i] + PHI*log(fabs(ARGUMENT_F1[i]));
            }
            else
            {
                RESULT[i] = XI1*ARGUMENT_F1[i];
            }
        }


        /*****************************************************************/
        /** 2. CALC GRADIENT                                            **/
        /*****************************************************************/
        if(nlhs==2)
        {
            /* DERIV_F1_DIAG = XI1 + bsxfun(@rdivide,PHI,Z) */
            if (PHI == 0.0)
            {
                for(i=0;i<M;i++) { DERIV_F1_DIAG[i] = XI1; }
            }
            else
            {
                for(i=0;i<M;i++) { DERIV_F1_DIAG[i] = XI1 + PHI/ARGUMENT_F1[NUM_O*M+i]; }
            }

            /* GRADIENT.LAMBDA(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG,bsxfun(@times,FVAL_F2,LOG_APOSTPDF_VALS{NUM_O}) * WEIGHTS' / sum(WEIGHTS) ...
                    - ARGUMENT_F1 * (LOG_APOSTPDF_VALS{NUM_O}*WEIGHTS'/sum(WEIGHTS))); */
            for(i = 0;i < K; i++)
            {
                daxpy(&M, &dMINUS_ONE, ARGUMENT_F1+NUM_O*M, &iONE, TMP_MATRIX+i*M, &iONE);
            }
            memcpy(TMP_WEIGHTS, WEIGHTS, K*sizeof(double));
            for(i = 0;i < K; i++) { TMP_WEIGHTS[i] *= aLOG_APOSTPDF_VALS[i]; }
            dgemv(chn, &M, &K, &dONE, TMP_MATRIX, &M, TMP_WEIGHTS, &iONE, &dZERO, GRAD_LAMBDA+NUM_O*M, &iONE );
            for(i=0;i<M;i++) { GRAD_LAMBDA[NUM_O*M+i] *= DERIV_F1_DIAG[i]; }

            /* GRADIENT.XI2(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG, ...
                       APOSTPDF_SAMPLES{NUM_O} * WEIGHTS' / sum(WEIGHTS) ); */
            dgemv(chn, &M, &K, &dONE, aAPOSTPDF_SAMPLES, &M, WEIGHTS, &iONE, &dZERO, GRAD_XI2+NUM_O*M, &iONE );
            for(i=0;i<M;i++) { GRAD_XI2[NUM_O*M+i] *= DERIV_F1_DIAG[i]; }


            /* GRADIENT.XI3(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG, ...
                       exp(bsxfun(@times,FPARAMS.PSI,Z)) * WEIGHTS' / sum(WEIGHTS) ); */
            dgemv(chn, &M, &K, &dONE, TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES, &M, WEIGHTS, &iONE, &dZERO, GRAD_XI3+NUM_O*M, &iONE );
            for(i=0;i<M;i++) { GRAD_XI3[NUM_O*M+i] *= DERIV_F1_DIAG[i]; }


            /* GRADIENT.PHI(:,NUM_O) = log(ARGUMENT_F1); */
            for(i=0;i<M;i++) { GRAD_PHI[NUM_O*M+i] = log(fabs(ARGUMENT_F1[M*NUM_O+i])); }

            /* GRADIENT.PSI(:,NUM_O) = bsxfun(@times,DERIV_F1_DIAG,...
                  (exp(bsxfun(@times,FPARAMS.PSI,APOSTPDF_SAMPLES{NUM_O})).*APOSTPDF_SAMPLES{NUM_O})*WEIGHTS' / sum(WEIGHTS)); */
            if (XI2 != 1.0)
            {
                for(i = 0;i < K*M; i++)
                {
                    TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES[i] *= aAPOSTPDF_SAMPLES[i];
                }
                dgemv(chn, &M, &K, &dONE, TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES, &M, WEIGHTS, &iONE, &dZERO, GRAD_PSI+NUM_O*M, &iONE );
                for(i=0;i<M;i++) { GRAD_PSI[NUM_O*M+i] *= XI3*DERIV_F1_DIAG[i]; }
            }
            else
            {
                for(i=0;i<M;i++) { GRAD_PSI[NUM_O*M+i] = 0; }
            }

        }



    }

    if (nlhs == 2) { free(DERIV_F1_DIAG); }

    free(WEIGHTS);
    free(TMP_WEIGHTS);
    free(TMP_MATRIX);
    free(TMP_MATRIX_EXP_PSI_aAPOSTPDFSAMPLES);

    }

    plhs[0] = mxRESULT;
    if (nlhs == 2)
    {
        plhs[1] = mxRESULT_GRAD;
    }
    else
    {
        free(ARGUMENT_F1);
    }

    return;
}
