function [FPARAMS,FVAL] = learn_estimator_family(JOINTPDF_SAMPLES, FUN_LOSSFUNCTION, APOSTPDF_SAMPLES, LOG_APOSTPDF_VALS, LEARNING_PARAMS)

M = size(APOSTPDF_SAMPLES{1},1); % M = number of unknown parameters

if isfield(LEARNING_PARAMS, 'PLOT_FUNCTION')
    global BEST_FVAL
    global BEST_FPARAMS
    BEST_FVAL = inf;
    BEST_FPARAMS = [];
end

%% DEFINE START - VALUES: corresponds approx. to MMSE estimator
LBOUND.XI1 = -inf;
LBOUND.XI2 = -inf;
LBOUND.XI3 = -inf;
LBOUND.LAMBDA = 0;

UBOUND.XI1 = inf;
UBOUND.XI2 = inf;
UBOUND.XI3 = inf;
UBOUND.LAMBDA = inf;


if LEARNING_PARAMS.UNIFORM == 1
    LBOUND.PHI = -inf;
    LBOUND.PSI = -inf;
    UBOUND.PHI = inf;
    UBOUND.PSI = inf;
else
    LBOUND.PHI = -inf(M,1);
    LBOUND.PSI = -inf(M,1);
    UBOUND.PHI = inf(M,1);
    UBOUND.PSI = inf(M,1);
end


% set up optimization problem
problem.objective = @internal_wrapper;
problem.x0        = struct2vect(LEARNING_PARAMS.FPARAMS_START);

problem.Aineq = [];problem.bineq = [];
problem.Aeq = [];problem.beq = [];
problem.nonlcon = [];

problem.lb = struct2vect(LBOUND);
problem.ub = struct2vect(UBOUND);

problem.solver = 'fmincon';
problem.options = optimset('Display','iter',...
    'TolX',LEARNING_PARAMS.TOLERANCES,'TolFun',LEARNING_PARAMS.TOLERANCES,...
    'MaxIter',LEARNING_PARAMS.MaxIter,'MaxFunEvals',LEARNING_PARAMS.MaxFunEvals);

if LEARNING_PARAMS.USE_GRADIENT == 1
    problem.options.GradObj = 'on';
    if LEARNING_PARAMS.DEBUG == 1
        problem.options.DerivativeCheck = 'on';
    end
end

[OPT_VALS,FVAL] = fmincon(problem);
FPARAMS = vect2struct(OPT_VALS);
fprintf('\n\n *** Stopped optimization with FVAL = %e\n\n',FVAL)

%% Subfunction wrapper()
    function [fval,grad] = internal_wrapper(x)
        if LEARNING_PARAMS.DEBUG == 1
            fprintf('DEBUG: starting internal wrapper with xi1 = %e,xi2 = %e,xi3 = %e,lambda = %e,phi = %e,psi = %e\n',...
                x(1),x(2),x(3),x(4),x(5),x(6))
        end
        if nargout == 1
            THETA_HAT = estimator_family(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS,vect2struct(x),LEARNING_PARAMS);
            LOSS = feval(FUN_LOSSFUNCTION,THETA_HAT,JOINTPDF_SAMPLES.THETA);
            fval = mean(LOSS);
            if imag(fval) ~= 0 || isnan(fval)
                disp(fval)
                fval = 1e9;
                fprintf('WARNING: Complex/NaN value of fval occured for parameter setting \n')
                vect2struct(x)
            end
        else
            [THETA_HAT, GRADIENT] = estimator_family(APOSTPDF_SAMPLES,LOG_APOSTPDF_VALS,vect2struct(x),LEARNING_PARAMS);
            [LOSS, GRADIENT_LOSS] = feval(FUN_LOSSFUNCTION,THETA_HAT,JOINTPDF_SAMPLES.THETA);
            fval = mean(LOSS);
            G.XI1 = mean(sum(GRADIENT.XI1 .* GRADIENT_LOSS,1));
            G.XI2 = mean(sum(GRADIENT.XI2 .* GRADIENT_LOSS,1));
            G.XI3 = mean(sum(GRADIENT.XI3 .* GRADIENT_LOSS,1));
            G.LAMBDA = mean(sum(GRADIENT.LAMBDA .* GRADIENT_LOSS,1));
            G.PHI = mean(sum(GRADIENT.PHI .* GRADIENT_LOSS,1));
            G.PSI = mean(sum(GRADIENT.PSI .* GRADIENT_LOSS,1));
            grad = struct2vect(G);            
            if any(imag([fval;grad]) ~= 0) || any(isnan([fval;grad]) ~= 0)
                disp(fval),disp(grad')
                fval = 1e9;
                grad(isnan(grad)) = eps;
                grad(imag(grad)~=0) = eps;
                fprintf('WARNING: Complex/NaN value of fval and/or grad occured for parameter setting \n')
                vect2struct(x)
            end
        end
        if LEARNING_PARAMS.DEBUG == 1
            fprintf('DEBUG: fval = %e\n',fval)
        end
        
        if LEARNING_PARAMS.DEBUG == 1 && nargout == 2 && LEARNING_PARAMS.CHECK_LEF_GRADIENT == 1
            fprintf('DEBUG: Comparison of Bayes risk gradients:\n')
            addpath('/home/m109/work/Matlab/Tools/DERIVESTsuite')
            LEARNING_PARAMS.DEBUG = 0;
            GRADESTI = gradest(@internal_wrapper, x);
            disp(grad(:)')
            disp(GRADESTI)
            LEARNING_PARAMS.DEBUG = 1;
            pause
        end
        
        if isfield(LEARNING_PARAMS, 'PLOT_FUNCTION' )
            if ~isnan(fval) && ~isinf(fval) && fval < BEST_FVAL
                BEST_FVAL = fval;
                BEST_FPARAMS = vect2struct(x);
                feval(LEARNING_PARAMS.PLOT_FUNCTION, THETA_HAT, APOSTPDF_SAMPLES, JOINTPDF_SAMPLES);
            end
        end
    end

end