% generate sample from truncated Gaussian
%
% Algorithm taken from
%
%   C. P. Robert: "Simulation of truncated normal variables",
%   Statistics and Computing, volume 5, number 2, pages 121-125, 1995

function samples = trunc_gaussian_vect(mu_minus,mu_plus,mu,variance,num_samples)

samples = zeros(1,num_samples);

not_found_sample = 1:num_samples;

mu_minus = (mu_minus - mu)/sqrt(variance);
mu_plus = (mu_plus - mu)/sqrt(variance);
interval = mu_plus - mu_minus;

while ~isempty(not_found_sample)
    
    z = rand(1,length(not_found_sample)) * interval + mu_minus;
    
    if mu_plus < 0
        
        rho = exp(1/2*(mu_plus^2 - z.^2));
        
    elseif mu_minus > 0
        
        rho = exp(1/2*(mu_minus^2 - z.^2));
        
    else
        
        rho = exp(-1/2*z.^2);
        
    end
    
    indices = rand(1,length(not_found_sample)) <= rho;
    samples(not_found_sample(indices)) = z(indices);
    not_found_sample(indices) = [];
end

samples = sqrt(variance) * samples + mu;

end